#include "camera.h"

#include <opencv/cv.h>
#include <opencv/highgui.h>

#include <iostream>

using namespace cv;
using namespace std;

// based on code provided at:
// http://www.aishack.in/2010/07/calibrating-undistorting-with-opencv-in-c-oh-yeah/

Camera::Camera()
{

	capture = VideoCapture(0);
	
	if(!capture.isOpened())
	{
		cerr << "Camera::Camera():  Camera capture device could not be opened." << endl;
		exit(1);
	}
	
}

Camera::~Camera()
{
	if(capture.isOpened())
	{
		capture.release();
	}
}

// returns a frame without compensating for lens/perspective distortion
Mat Camera::getRawFrame()
{
	Mat frame;
	capture >> frame;
	return frame;
}

// returns a frame that has been corrected for distortion - this should
// only be called after a successful calibration
Mat Camera::getUndistortedFrame()
{

	Mat undistorted;
	Mat frame = getRawFrame();
	
	undistort(frame, undistorted, intrinsic, distCoeffs);
	
	return undistorted;

}

// performs the calibration process, either if (a) a previous calibration settings
// file could not be found, or (b) if the user asks for one
void Camera::beginCalibration()
{

	const char *WINDOW_NAME = "cal";
	const char *CALIBRATION_FILENAME = "calibration.yml";

	int numBoards = 0;				// number of board captures
	int numHorizontalCorners;		// internal corners
	int numVerticalCorners;
	
	int numSquares;					// total number of squares
	Size boardSize;
	
	vector<vector<Point3f> > objectPoints;
	vector<vector<Point2f> > imagePoints;
	vector<Point2f> corners;
	vector<Point3f> obj;
	int captureSuccesses = 0;
	
	Mat image;
	Mat greyImage;
	bool found;
	int key;

	vector<Mat> rvecs;
	vector<Mat> tvecs;
	
	char ch;
	bool useCalibrationFile = false;
	
	FileStorage io;
	
	cout << "Camera::beginCalibration(): use calibration file? y/n: ";
	cin >> ch;
	
	if(ch == 'y')
	{

		useCalibrationFile = true;
		io = FileStorage(CALIBRATION_FILENAME, FileStorage::READ);
		io["intrinsic"] >> intrinsic;
		io["distCoeffs"] >> distCoeffs;
		
		io.release();
		
		if(!intrinsic.data || !distCoeffs.data)
		{
			useCalibrationFile = false;
			cerr << "Camera::beginCalibration(): calibration file(s) not found" << endl;
		}
		else
		{
			cout << "Camera::beginCalibration(): files found; calibration successful" << endl;
		}
		
	}
	
	if(!useCalibrationFile)
	{
	
		cout << "Camera::beginCalibration(): Beginning manual calibration" << endl;
	
		cout << " Enter number of internal corners along horizontal: ";
		cin >> numHorizontalCorners;
		
		cout << " Enter number of internal corners along vertical: ";
		cin >> numVerticalCorners;
		
		cout << " Enter number of boards to capture: ";
		cin >> numBoards;
		
		numSquares = numHorizontalCorners * numVerticalCorners;
		boardSize = Size(numHorizontalCorners, numVerticalCorners);
		
		image = getRawFrame();
		namedWindow(WINDOW_NAME);
		
		// CHANGE THIS UNIT SYSTEM LATER
		for(int i = 0; i < numSquares; i ++)
		{
			obj.push_back(Point3f(i / numHorizontalCorners, i % numHorizontalCorners, 0.0));
		}
		
		while(captureSuccesses < numBoards)
		{
		
			cvtColor(image, greyImage, CV_BGR2GRAY);
			
			found = findChessboardCorners(image, boardSize, corners, CV_CALIB_CB_ADAPTIVE_THRESH | CV_CALIB_CB_FILTER_QUADS);
			if(found)
			{
				cornerSubPix(greyImage, corners, Size(11, 11), Size(-1, -1), TermCriteria(CV_TERMCRIT_EPS | CV_TERMCRIT_ITER, 30, 0.1));
				drawChessboardCorners(greyImage, boardSize, corners, found);
			}
			
			imshow(WINDOW_NAME, greyImage);
			
			image = getRawFrame();
			
			key = waitKey(1);
			
			if(key == 27)
			{
				return;
			}
			else if(key == ' ' && found)
			{
			
				imagePoints.push_back(corners);
				objectPoints.push_back(obj);
			
				captureSuccesses ++;
				cout << "Camera::beginCalibration(): capture " << captureSuccesses << "/" << numBoards << " stored" << endl;
				
			}
		
		}
		
		intrinsic = Mat(3, 3, CV_32FC1);
		intrinsic.ptr<float>(0)[0] = 1;
		intrinsic.ptr<float>(1)[1] = 1;
		
		calibrateCamera(objectPoints, imagePoints, image.size(), intrinsic, distCoeffs, rvecs, tvecs);
		
		io = FileStorage(CALIBRATION_FILENAME, FileStorage::WRITE);
		io << "intrinsic" << intrinsic;
		io << "distCoeffs" << distCoeffs;
		io.release();
		
		cout << "Camera::beginCalibration(): Calibration complete and saved to file" << endl;
		
		destroyWindow(WINDOW_NAME);
	
	}

}
