#include "util.h"
#include "camera.h"
#include "formation.h"
#include "robotcomm.h"

#include "pawprint.h"
#include "pawprinttracker.h"

#include <opencv/cv.h>
#include <opencv/highgui.h>

#include <iostream>
#include <vector>

using namespace cv;
using namespace std;

// Serial communications code from
// http://www.codeguru.com/cpp/i-n/network/serialcommunications/article.php/c2503/CSerial--A-C-Class-for-Serial-Communications.htm

// NOTE: ensure that your PC's Wixel is on any of COM ports 1 through 4 - the serial comms
// code will not work otherwise

void overlayFrameData(Mat &frame, vector<PawPrint*> &robotPrints, vector<Point2f> robotPoints, int formation)
{

	const Point2f FORMATION_TEXT_POSITION(10, 20);
	const Scalar FORMATION_TEXT_COLOR(255, 255, 255);
	const float FORMATION_TEXT_SCALE = 0.4;

	const int BRIGHTNESS = 190;
	const Scalar FORMATION_COLORS[] = {Scalar(0, 0, BRIGHTNESS), Scalar(0, BRIGHTNESS, 0), Scalar(0, BRIGHTNESS, BRIGHTNESS),
									   Scalar(BRIGHTNESS, 0, 0), Scalar(BRIGHTNESS, 0, BRIGHTNESS), Scalar(BRIGHTNESS, BRIGHTNESS, 0)};
	int i;
	
	// pass target information onto the robots
	for(i = 0; i < (int)robotPrints.size(); i ++)
	{
		circle(frame, robotPoints[i], 3, FORMATION_COLORS[i]);
		circle(frame, robotPrints[i] -> getCenter(), 3, FORMATION_COLORS[i]);
	}
	
	// print out formation
	if(formation == Formation::VERTICAL_LINE)
	{
		putText(frame, "Formation: Vertical line", FORMATION_TEXT_POSITION, FONT_HERSHEY_SIMPLEX, FORMATION_TEXT_SCALE, FORMATION_TEXT_COLOR);
	}
	else if(formation == Formation::DIAGONAL_LINE)
	{
		putText(frame, "Formation: Diagonal line", FORMATION_TEXT_POSITION, FONT_HERSHEY_SIMPLEX, FORMATION_TEXT_SCALE, FORMATION_TEXT_COLOR);
	}
	else if(formation == Formation::HORIZONTAL_LINE)
	{
		putText(frame, "Formation: Horizontal line", FORMATION_TEXT_POSITION, FONT_HERSHEY_SIMPLEX, FORMATION_TEXT_SCALE, FORMATION_TEXT_COLOR);
	}
	else
	{
		putText(frame, "Formation: Circle", FORMATION_TEXT_POSITION, FONT_HERSHEY_SIMPLEX, FORMATION_TEXT_SCALE, FORMATION_TEXT_COLOR);
	}

}

int main(int args, char *argv[])
{
	
	const float FRAME_SCALE_FACTOR = 0.75;
	const char* WINDOW_TRACKING = "Tracking Summary";
    
	Camera *camera;							// our camera feed
	
	Mat frame;								// color capture frame
	Mat frame_gray;							// greyscale capture frame

    vector<PawPrint*> pawprints;			// unsorted tracking locations for each robot
    vector<Point2f> formationPoints;		// unsorted list of formation points
    vector<Point2f> robotPoints;			// sorted target formation locations for each robot
	vector<PawPrint*> robotPrints;			// sorted tracking locations for each robot
	vector<Point2f>::iterator i;
	
	Point2f average;
	int printIndex;
	int pointIndex;
	
	int formation = Formation::nextFormation();
    int key;
    
    // print intro
    cout << " - - - Paw Print Vision Server - - -" << endl;
    cout << "- - - - - - - - - - - - - - - - - -\n" << endl;
    
    // setup our camera
    camera = new Camera();
    
    // uncomment for camera calibration
    //camera -> beginCalibration();
    
    // setup the paw print tracking
    PawPrintTracker::init();
    PawPrintTracker::openTrackingWindow();
	
	// setup window
	namedWindow(WINDOW_TRACKING, 1);
	
	// initialize serial comms
	RobotComm::init(1);
	
	// while the user hasn't pressed the escape key...
	do
	{
	
		// get a frame from our calibration camera
		//frame = camera -> getUndistortedFrame();	// uncomment and swap for below line for undistorted images
		frame = camera -> getRawFrame();
		resize(frame, frame, Size(), FRAME_SCALE_FACTOR, FRAME_SCALE_FACTOR);
		
		// perform paw print tracking
		PawPrintTracker::preprocessFrame(frame, frame_gray);
		PawPrintTracker::findPawPrints(frame_gray, pawprints);
		
		// darken our frame so we can see the pawprints and robot overlays better
		frame *= 0.4;
		PawPrintTracker::drawPawPrints(frame, pawprints);
		
		// update our pawprint tracking window
		PawPrintTracker::updateTrackingWindow();
		
		// determine the points that need to be occupied by the robots
		Formation::getFormationPoints((int)pawprints.size(), formationPoints, frame.size());
		
		// clear the sorted robot tracking data
		robotPrints.clear();
		robotPoints.clear();
		
		// assign each robot to a formation point
		while(formationPoints.size() > 0)
		{
		
			// compute average position of formation points
			average = Util::average(formationPoints);
			
			// find the robot furthest away from this average
			printIndex = Formation::getFurthestPrintFromPoint(average, pawprints);
			
			// find the point closest to this robot
			pointIndex = Formation::getClosestPointToPrint(formationPoints, pawprints[printIndex]);
			
			// put the computed print and formation point in a parallel array
			robotPoints.push_back(formationPoints[pointIndex]);
			robotPrints.push_back(pawprints[printIndex]);
			
			// remove the print and the point so they are not considered in future calculations
			formationPoints.erase(formationPoints.begin() + pointIndex);
			pawprints.erase(pawprints.begin() + printIndex);
		
		}
		
		// render formation and robot data to the frame
		overlayFrameData(frame, robotPrints, robotPoints, formation);
		
		// send targeting data to robots
		RobotComm::controlComms(robotPoints, robotPrints);
		
		// grab a keypress from the user
		key = waitKey(30);
		if(key == 'f')			// change formation
		{
			formation = Formation::nextFormation();
		}
		
		// show tracking summary
		imshow(WINDOW_TRACKING, frame);
	
	// exit if the user presses ESC
	} while(key != 27);
	
	// free up our memory
	RobotComm::kill();
	delete camera;

	return 0;

}
