#include "pawprinttracker.h"
#include "pawprint.h"
#include "util.h"

#include <climits>
#include <iostream>
#include <opencv/cv.h>
#include <vector>

using namespace cv;
using namespace std;

const char *PawPrintTracker::TRACKING_WINDOW_NAME = "PawPrintTracker Window";

Mat PawPrintTracker::trackingWindowFrame;

int PawPrintTracker::thresholdValue;
int PawPrintTracker::erodeValue;
int PawPrintTracker::dilateValue;

SimpleBlobDetector *PawPrintTracker::blobDetector = NULL;

void PawPrintTracker::init()
{
	thresholdValue = 63;
	erodeValue = 1;
	dilateValue = 2;
	initBlobDetectors();
}

void PawPrintTracker::openTrackingWindow()
{
	namedWindow(TRACKING_WINDOW_NAME, 1);
	cvCreateTrackbar("threshold", TRACKING_WINDOW_NAME, &thresholdValue, 255, NULL);
	cvCreateTrackbar("erode", TRACKING_WINDOW_NAME, &erodeValue, 15, NULL);
	cvCreateTrackbar("dilate", TRACKING_WINDOW_NAME, &dilateValue, 15, NULL);
}

void PawPrintTracker::updateTrackingWindow()
{
	imshow(TRACKING_WINDOW_NAME, trackingWindowFrame);
}

void PawPrintTracker::preprocessFrame(Mat frame, Mat &result)
{

	result = frame;

	cvtColor(result, result, CV_BGR2GRAY);
	threshold(result, result, thresholdValue, 255, THRESH_BINARY_INV);
	erode(result, result, Mat(Size(erodeValue, erodeValue), CV_8UC1));
	dilate(result, result, Mat(Size(dilateValue, dilateValue), CV_8UC1));
	
}

void PawPrintTracker::findPawPrints(Mat &frame, vector<PawPrint*> &prints)
{
	
	vector<KeyPoint> keypoints;
	vector<KeyPoint> centers;
	prints.clear();
	
	findPawCenters(frame, keypoints, centers);
	drawKeypoints(frame, keypoints, frame, Scalar(0, 0, 255));
	findPawDigits(frame, keypoints, centers, prints);
	
	trackingWindowFrame = frame;

}

void PawPrintTracker::drawPawPrints(Mat &frame, vector<PawPrint*> &prints)
{

	vector<PawPrint*>::iterator i;
	
	for(i = prints.begin(); i != prints.end(); i ++)
	{
		(*i) -> draw(frame);
	}

}

void PawPrintTracker::initBlobDetectors()
{

	SimpleBlobDetector::Params params;
	
	params.filterByArea = true;
	params.minArea = 1;
	params.maxArea = 1500;
	params.minDistBetweenBlobs = 1;
	params.filterByColor = true;
	params.blobColor = 255;

	params.filterByConvexity = true;
	params.minConvexity = 0.7;//0.8;//0.85;
	params.maxConvexity = 1.01;

	params.filterByCircularity = true;
	params.minCircularity = 0.7;//0.8;//0.83;
	params.maxCircularity = 1.01;

	blobDetector = new SimpleBlobDetector(params);
	blobDetector -> create("BlobDetector");

}

void PawPrintTracker::findPawCenters(Mat &frame, vector<KeyPoint> &keypoints, vector<KeyPoint> &centers)
{

	const float MIN_TOLERANCE = 0.8;
	const float MAX_TOLERANCE = 1.2;

	vector<KeyPoint>::iterator i;
	KeyPoint curr;
	float size;
	bool done = false;

	// detect all blobs and sort then in non-ascending order
	blobDetector -> detect(frame, keypoints);
	Util::sortKeyPointsDownBySize(keypoints);
	
	// grab the largest n similar blobs and create pawprints out of them
	if(keypoints.size() > 0)
	{

		// use the first large blob's size as a guideline
		size = keypoints[0].size;

		// continually grab large blobs until we find one that deviates from the first's size by too much
		i = keypoints.begin();
		while(!done && i != keypoints.end())
		{
		
			curr = *i;
			
			done = !(curr.size > size * MIN_TOLERANCE && curr.size < size * MAX_TOLERANCE);
			if(!done)
			{
				centers.push_back(*i);
			}
			
			i ++;
			
		}
	
	}

}

void PawPrintTracker::findPawDigits(Mat &frame, vector<KeyPoint> &keypoints, vector<KeyPoint> &centers, vector<PawPrint*> &prints)
{

	const int MAX_ROBOT_ID = 32;
	const float MAX_DIST_TOLERANCE = 2.0;

	vector<KeyPoint>::iterator i, j;
	vector<Point2f>::iterator k;
	vector<Point2f> digits;
	
	KeyPoint center;
	Point2f potentialDigit;

	float maxDistanceFromCenter;
	
	if(keypoints.size() > 0)
	{
	
		maxDistanceFromCenter = keypoints[0].size * MAX_DIST_TOLERANCE;
	
		for(i = centers.begin(); i != centers.end(); i ++)
		{
		
			center = *i;
			digits.clear();
			
			if(getSmallestDigit(keypoints, center.pt, maxDistanceFromCenter, potentialDigit))
			{
				digits.push_back(potentialDigit);
			}
			
			// if this center has one digit around it, it is a paw print;
			// this is the only place where 1 digit is enforced, so it can be
			// changed in the future if required
			if(digits.size() == 1)
			{
				int id = getPrintID(center.pt, digits[0], frame);
				if(id <= MAX_ROBOT_ID)
				{
					prints.push_back(new PawPrint(center.pt, digits[0], id));
				}
			}
		
		}
	
	}

}

bool PawPrintTracker::getSmallestDigit(vector<KeyPoint> &keypoints, Point2f center, int maxDist, Point2f &result)
{

	vector<KeyPoint>::iterator i;
	int smallestSize = INT_MAX;
	float dist;
	bool found = false;
	
	for(i = keypoints.begin(); i != keypoints.end(); i ++)
	{
	
		dist = Util::length(i -> pt - center);
		if(dist < maxDist && dist > 1 && i -> size < smallestSize)
		{
			smallestSize = i -> size;
			result = i -> pt;
			found = true;
		}
	
	}
	
	return found;

}

int PawPrintTracker::getPrintID(Point2f center, Point2f digit, Mat &frame)
{

	const int WEDGE_COUNT = 8;
	const float WEDGE_SIZE = 360.0 / (float)WEDGE_COUNT;
	const int WEDGE_SAMPLE_OFFSET = Util::length(center - digit);
	
	Point2f sample;
	bool value;
	
	int id = 0;
	int angle = (360.0 - Util::angle(center - digit)) - (WEDGE_SIZE * 1.5);
	int i = 0;
	
	while(i < (WEDGE_COUNT - 2))
	{
	
		sample = center + Util::vecAtAngle(angle) * WEDGE_SAMPLE_OFFSET;
		
		// ensure we don't go past the bounds of the image
		if(sample.x > 0 && sample.y > 0 && sample.x < frame.size().width && sample.y < frame.size().height)
		{
			value = (int)frame.at<Vec3b>(sample)[0] > 0;
			id += (int)value * pow(2, i);
		}
		
		angle -= WEDGE_SIZE;
		i ++;
	
	}
	
	return id;

}
