#include "robotcomm.h"
#include "pawprint.h"
#include "util.h"

#include <vector>
#include <opencv/cv.h>

using namespace std;
using namespace cv;

bool RobotComm::initialized = false;

CSerial RobotComm::serial;

void RobotComm::init(int port)
{

	// setup serial connection
    // ** NOTE: only COM port values 1-4 are valid **
    
    // (see http://www.codeguru.com/cpp/i-n/network/serialcommunications/
    // article.php/c2503/CSerial--A-C-Class-for-Serial-Communications.htm
    // for more details)
	serial.Open(port, 9600);
	if(!serial.IsOpened())
	{
		cout << "RobotComm::init(): could not open serial port " << port << endl;
	}
	else
	{
		initialized = true;
	}

}

void RobotComm::kill()
{
	serial.Close();
	initialized = false;
}

void RobotComm::controlComms(vector<Point2f> &robotPoints, vector<PawPrint*> &robotPrints)
{

	char data;
	int i;
	
	if(initialized)
	{
	
		// pass target information onto the robots
		for(i = 0; i < (int)robotPrints.size(); i ++)
		{
			
			// send new channel data
			data = 100 + (robotPrints[i] -> getID() * 10);
			serial.SendData(&data, 1);
		
			// wait until the Wixel has changed channel
			while(serial.ReadDataWaiting() < 1);
			serial.ReadData((void*)&data, 1);
			
			// only proceed if we receive confirmation that the channel change was successful
			if(data)
			{
			
				// send target information
				data = getTargetByte(robotPrints[i], robotPoints[i]);
				serial.SendData(&data, 1);
				
				// wait until we receive confirmation that the Wixel has transmitted data
				while(serial.ReadDataWaiting() < 1);
				serial.ReadData((void*)&data, 1);
			
			}
			else
			{
				cout << "RobotComms::controlComms(): Could not link to robot ID " << robotPrints[i] -> getID() << endl;
			}
		
		}
	
	}

}

char RobotComm::getTargetByte(PawPrint *print, Point2f point)
{

	// we compress a robot's target's distance and relative angle into
	// a single byte, with the first 2 bits representing distance, and
	// the last 6 bits representing the angle to the target
	
	// [xx]    [xxxxxx]
	// [dist]  [angle]
	
	// since we use two bytes for distance, we say that 0 is "at destination",
	// 1 is "close to destination", and 2 is "far away"
	
	Point2f relative = print -> getCenter() - point;
	Point2f relativeVec = Util::normalize(relative);
	Point2f orientationVec = Util::normalize(print -> getOrientation());
	char data;

	float targetAngleDiff;
	float targetDistDiff;
	
	// wrap the target angle difference from -180 to 180
	targetAngleDiff = Util::wrapvalue(Util::angle(relativeVec) - Util::angle(orientationVec));
	targetAngleDiff -= 180.0;
	
	// get the distance to the target
	targetDistDiff = Util::length(relative);
	
	// indicate speed of movement based on distance to target
	if(targetDistDiff < 10)
		data = 0;
	else if(targetDistDiff < 50)
		data = 1 << 6;
	else
		data = 2 << 6;
	
	// we can still get 6 degrees of precision in our robot's angle with 6 bits
	data |= (int8_t)((targetAngleDiff + 180) / 6);
	
	return data;

}
